// Datei LiesOhneReturn.c
/* ------------------------------------------------------------------------
Wenn man Daten von der Tastatur einliest, blockiert ("wartet") der
Lesebefehl normalerweise so lange, bis der Benutzer auf die Return-Taste
drueckt.

Dieses Programm zeigt, wie ein Programm Zeichen sofort (engl. immediately)
von der Tastatur lesen kann, d.h. OHNE darauf zu warten, dass der Benutzer,
auf die Return-Taste drueckt.

Ausserdem zeigt dieses Programm, wie man alle sofort verfuegbaren Zeichen
einlesen kann (ohne auf weitere Eingaben des Benutzers zu warten). Bei
einem solchen Lesevorgang werden eventuell 0 Zeichen eingelesen.

Die hier verwendeten Befehle sind nicht standardisiert, sondern abhaengig
vom verwendeten Betriebssystem. Sie funktionieren z.B. unter Suse-Linux
Version 10.0 in einem Konsolen- oder einem Xterm-Fenster, einigermassen
auch in einem Cygwin-bash-Fenster unter Windows und auf einigen weiteren
Unix-Systemen).
------------------------------------------------------------------------
Alle geoeffneten Dateien eines Programms sind mit 0 beginnend
durchnummeriert. Dabei bezeichnet (unter Unix-Betriebssystemen) die
Nummer 0 immer die Standardeingabe, die schon beim Start des Programms
geoeffnet und normalerweise mit der Tastatur verbunden ist.
------------------------------------------------------------------------ */
#include <stdio.h>   // E/A in Bytestroeme, printf, scanf, sprintf, ...
#include <stdlib.h>  // system
#include <stdbool.h> // true, false, bool
#include <fcntl.h>   // fcntl
#include <ctype.h>   // isprint
#include <string.h>  // strlen
// ------------------------------------------------------------------------
int setNonblockFlag(int dateiNr, int neuerWert) {
   // Das NONBLOCK-Bit der Standardeingabe wird auf den durch neuerWert
   // bezeichneten Wert (0 bzw. 1) gesetzt. Wenn das Bit gleich 0 ist,
   // blockiert ("wartet") ein Lesebefehl wie z.B. getchar so lange,
   // bis ein Zeichen gelesen werden kann und liest es. Wenn das
   // NONBLOCK-Bit auf 1 steht, blockiert ein Lesebefehl nicht, sondern
   // kehrt sofort zurueckt (evtl. ohne ein Zeichen gelesen zu haben).

   // Steuerbits (flags) der Standardeingabe nach steuerBits holen:
   int steuerBits = fcntl(dateiNr, F_GETFL, 0);
   if (steuerBits == -1) {
      printf("Steuerbits der Datei Nr. %d konnten "
             "nicht gelesen werden!\n", dateiNr);
      return -1;
   }

   // Jetzt wird das NONBLOCK Bit in steuerBits auf 1 bzw. 0 gesetzt:
   if (neuerWert != 0) {
      steuerBits |=  O_NONBLOCK; // Bit auf 1 setzen
   } else {
      steuerBits &= ~O_NONBLOCK; // Bit auf 0 setzen
   }
   // Steuerbits (flags) der Standardeingabe zurueckschreiben:
   return fcntl(dateiNr, F_SETFL, steuerBits);
}
// ------------------------------------------------------------------------
char * liesAllesOhneZuWarten(int max, char * puffer) {
   // Blockiert ("wartet") bis von der Standardeingabe mindestens
   // 1 Zeichen gelesen werden kann. Liest dann alle in diesem
   // Moment lesbaren Zeichen (hoechstens aber max viele) nach
   // puffer und liefert puffer als Ergebnis
   // Der puffer enthaelt dann mindestens 1 Zeichen, hoechstens aber
   // max viele Zeichen.

   int i = 0;
   // Auf das 1. Zeichen wird notfalls gewartet:
   int naechstesZeichen = getchar();

   // Das NONBLOCK-Bit der Standardeingabe auf 1 setzen (damit
   // Lesebefehle nicht mehr blockieren):
   setNonblockFlag(0, 1);

   // Weitere Zeichen einlesen bis getchar als Ergebnis -1 liefert
   // (d.h. bis kein Zeichen mehr sofort lesbar ist):
   while (true) {
      puffer[i++] = (char) naechstesZeichen;
      naechstesZeichen = getchar();
      if (naechstesZeichen < 0) break; // Kein Zeichen mehr lesbar
      if (i == max-1)           break; // puffer ist voll
   }
   // Abschliessende 0 nach puffer:
   puffer[i] = 0;

   // Das NONBLOCK-Bit der Standardeingabe wieder auf 0 setzen;
   setNonblockFlag(0, 0);
   return puffer;
}
// ------------------------------------------------------------------------
// Eine Konstante, die als Bestandteil von printf-Formatstrings
// benutzt wird (die konkrete Zahl 12 soll nur an einer Stelle des
// Programms stehen, damit sie leichter zu aendern ist):
#define F1 "12"
// ------------------------------------------------------------------------
void gibAusAlsZeichen(char * str) {
   // Gibt die Zeichen von str als einzelne Zeichen aus. Anstelle eines
   // nicht-druckbaren Zeichens wird ein Punkt . ausgegeben.
   int  i = 0;
   char c;
   printf("%-"F1"s: ", "Als Zeichen");
   while (true) {
      c = str[i++];
      if (c == 0) break;
      // Jedes nicht-druckbare Zeichen durch einen Punkt ersetzen:
      if (! isprint(c)) c = '.';
      printf("%3c ", c);     // c als Zeichen ausgeben
   }
   printf("\n");
}
// ------------------------------------------------------------------------
void gibAusHexadezimal(char * str) {
   // Gibt die Zeichen von str als Zahlen im 16-er-System (Hex-Zahlen)
   // aus:
   int  i = 0;
   unsigned char c;
   printf("%-"F1"s: ", "Hexadezimal");
   while (true) {
      c = str[i++];
      if (c == 0) break;
      printf(" %02X ", c);   // c hexadezimal ausgeben
   }
   printf("\n");
}
// ------------------------------------------------------------------------
void gibAusDezimal(char * str) {
   // Gibt die Zeichen von str als Zahlen im 10-er-System (Dezimalzahlen)
   // aus:
   int i = 0;
   unsigned u;
   printf("%-"F1"s: ", "Dezimal");
   while (true) {
      // Der Wert von str[i] wird explizit von char nach unsigned char
      // und dann implizit von unsigned char nach unsigned int umgewandelt:
      u = (unsigned char) str[i++];
      if (u == 0) break;
      printf("%3u ", u);            // u dezimal ausgeben
   }
   printf("\n");
}
// ------------------------------------------------------------------------
int main() {
   int c;

   // stty bezeichnet die Standardeingabe (das ist meistens die Tastatur).
   // Beim Einlesen von stty soll das Betriebssystem die Zeichen
   // sofort an dieses Programm geben, und nicht erst dann, wenn
   // der Benutzer auf die Return-Taste drueckt.
   // Dieser Befehl ist stark abhaengig vom verwendeten Betriebssystem
   // (und den besseren Befehl mit ioctl habe ich "verlegt" und kann ihn
   // gerade nicht wiederfinden):
   system("/bin/stty cbreak");

   // Von der Standardeingabe solange Zeichen einlesen und (in mehreren
   // Formen) zur Standardausgabe ausgeben bis der Benutzer ein q eingibt:
   int const   PLEN = 10; // Vermutlich wuerden 5 Zeichen genuegen, aber
   char puffer[PLEN];     // sicher ist sicher.

   while (true) {
      printf("-----------------------------------"
             "---------------------------------\n");

      printf("Ein Tastendruck (q zum Beenden): ");
      liesAllesOhneZuWarten(PLEN, puffer);
      // Wieviele Zeichen wurden tatsaechlich in den puffer gelesen?
      int          anz     = strlen(puffer);
      // Singular und Plural unterscheiden:
      const char * wurde_n = (anz==1) ? "wurde" : "wurden";
      printf("  Eingelesen %s %d Zeichen:\n", wurde_n, anz);

      gibAusAlsZeichen (puffer);
      gibAusHexadezimal(puffer);
      gibAusDezimal    (puffer);

      if (puffer[0] == 'q') break;
   }
   printf("-----------------------------------"
          "---------------------------------\n");
   printf("LiesOhneReturn beendet sich!\n");
   system("/bin/stty -cbreak");              // stty wieder "zuruecksetzen"

   return 0;
}
/* ------------------------------------------------------------------------
Ein Dialog mit diesem Programm (der Benutzer hat dabei auf folgende
Tasten gedrueckt: A, u-Umlaut, Pfeil nach oben, Bild nach unten, q):

--------------------------------------------------------------------
Ein Tastendruck (q zum Beenden): A  Eingelesen wurde 1 Zeichen:
Als Zeichen :   A
Hexadezimal :  41
Dezimal     :  65
--------------------------------------------------------------------
Ein Tastendruck (q zum Beenden): ü  Eingelesen wurden 2 Zeichen:
Als Zeichen :   .   .
Hexadezimal :  C3  BC
Dezimal     : 195 188
--------------------------------------------------------------------
Ein Tastendruck (q zum Beenden): ^[[A  Eingelesen wurden 3 Zeichen:
Als Zeichen :   .   [   A
Hexadezimal :  1B  5B  41
Dezimal     :  27  91  65
--------------------------------------------------------------------
Ein Tastendruck (q zum Beenden): ^[[6~  Eingelesen wurden 4 Zeichen:
Als Zeichen :   .   [   6   ~
Hexadezimal :  1B  5B  36  7E
Dezimal     :  27  91  54 126
--------------------------------------------------------------------
Ein Tastendruck (q zum Beenden): q  Eingelesen wurde 1 Zeichen:
Als Zeichen :   q
Hexadezimal :  71
Dezimal     : 113
--------------------------------------------------------------------
LiesOhneReturn beendet sich!
------------------------------------------------------------------------ */
